#
# encoding: utf-8
# api: fpm
# title: Composer source
# description: Converts composer bundles into system or/and phar packages
# type: package
# depends: bin:composer
# category: source
# version: 0.6
# state: beta
# license: MITL
# author: mario#include-once:org
# 
# Creates system packages for composer/packagist bundles.
#
#  → Either works on individual vendor/*/*/ paths as input
#    (the vendor/ prefix can be omitted).
#
#  → Or downloads a single vnd/pkgname bundle.
#
# Supports different target variations:
#
#  → Syspackages (deb/rpm) end up in /usr/share/php/Vnd/Pkg.
#
#  → Whereas --composer-phar creates Phars wrapped into system
#    packages, with target names of /usr/share/php/vnd-pkg.phar.
#
#  → While a standard `-t phar` target will just compact individual
#    components into localized phars.
#
# NOTES
#
# · The build process utilizes `composer require` to fetch new packages,
#   if xpm -s composer isn't run from within a composer managed project.
#
# · The vendor/ extraction prefix isn't retained any longer. Composer
#   wasn't meant to manage system-globally installed libraries.
#
# · System packages thus need a global autoloader (shared.phar / phpab)
#   or manual includes.
#
# · Dependencies are attempted to be Debian dh_phpcomposer/pkg-php compatible.
#   For RPM compatibility http://fedoraproject.org/wiki/Packaging:PHP is used.
#
# · Phar targets are intended as application-level plugins, not just plain
#   bundle dumps.
#    · Use --attr to override and add description and control fields.
#    · Can be combined with `-u lcase=php` filter to lowercase filenames
#      within → for PHP-compliant pathmap loaders.
#    · Planning: add classmap builder to --phar target.
#

require "fpm/package"
require "fpm/util"
require "fpm/errors"
require "fpm/package/zip"
require "fileutils"
require "json"


# Composer reading (source only)
class FPM::Package::Composer < FPM::Package

  option "--ver", "1.0\@dev", "Which version to checkout", :default=>nil
  option "--composer-phar", :flag, "Convert bundle into .phar wrapped in system package", :asis=>true, :default=>false
  option ["--phar", "--phar-composer"], :flag, "(alias)", :asis=>true, :hidden=>true, :attribute_name=>"composer_phar"
  option "--no-alt", :flag, "Do not allow phar- and php- alternatives.", :hidden=>true, :default=>false
  
  public

  def initialize(*args)
    super(*args)
    @architecture = "all"
    @name_prefix = ""      # hold "php-" or "phar-" syspackage name prefix
    @target_dir = nil      # two-level base directory under /usr/share/php
    @once = false          # prevent double .input() invocation
  end
  attr_reader :in_bundle

  # Download composer bundle, or compact from existing vendor/ checkout
  def input(vnd_pkg_path)

    # general params
    @in_bundle = vnd_pkg_path.gsub(/^(.+\/+)*vendor\/+|\/(?=\/)|\/+$/, "")
    @name = in_bundle.gsub(/[\W]+/, "-")
    json = {}
    if @once
      @once = true
      raise FPM::InvalidPackageConfiguration, "You can't input multiple bundle names. Only one package can be built at a time currently. Use a shell loop please."
    elsif in_bundle =~ /^composer\/\w+\.\w+/
      raise FPM::InvalidPackageConfiguration, "composer/*.* files specified as input. Supply only one bundle id."
    end

    # Copy existing vendor/*/*/ file tree
    if File.exist?("vendor/" + in_bundle)
      json = parse_lock("composer.lock", in_bundle)[in_bundle]
      # localize contents below vendor/*/*/ input directory
      ::Dir.chdir("./vendor/#{in_bundle}/#{json['target-dir']}/") do
        FileUtils.cp_r(glob("./*"), build_path)
      end
    else
      # Download one package (and dependencies, which are thrown away afterwards)
      ::Dir.chdir(staging_path) do
        ver = attributes[:composer_ver]
        safesystem(
          composer, "require", "--prefer-dist", "--update-no-dev", "--ignore-platform-reqs",
          "--no-ansi", "--no-interaction", in_bundle, *(ver ? [ver] : [])
        )
        # localize Vnd/Pkg folder
        json = parse_lock("composer.lock", in_bundle)[in_bundle]
        FileUtils.mv(glob("./vendor/#{in_bundle}/#{json['target-dir']}/*"), build_path)
      end
    end

    # At this point the build_path contains just the actual class files, etc.
    # Conversion to sys/phar/sysphar is handled in convert() along with the
    # dependency translation.
    composer_json_import(json)
    cleanup_staging
    # Defaults for packaging targets
    attributes[:prefix] ||= "/usr/share/php"
    @target_dir = json.include?("target-dir") ? json["target-dir"] : in_bundle
    attributes[:phar_format] = "zip+gz" unless attributes[:phar_format_given?]
  end # def output

  
  # Decorate files/includes with :prefix for target package type
  def convert(klass)
    # Subclass needs target-specific staging_path subdir (else may inherit cruft for multi-target building)
    typ = package_typ(klass)
    pkg = super(klass)
    pkg.staging_path = staging_path(klass.to_s.gsub(/\W/, '-'), :mkdir)

    # workaround for update-filter modules expecting to work on real staging_path
    if typ == :filter
      pkg.staging_path = @build_path

    # becomes local -t phar
    elsif typ == :phar
      pkg.staging_path = build_path
      @name_prefix = nil  # needs to be reset in case of multi-target building

    # prepare matroska phar-in-deb/rpm, ends up in /usr/share/php/*.phar
    elsif attributes[:composer_phar_given?]
      FileUtils.mkdir_p(staging_dest = pkg.staging_path + @attributes[:prefix])
      ::Dir.chdir("#{build_path}") do
        phar = convert(FPM::Package::Phar) # will loop internally
        phar.output("#{staging_dest}/#{@name}.phar");
      end
      @name_prefix = "phar-"
      pkg.provides << "php-#{@name}"

    # system package (deb/rpm/*) with plain files under /usr/share/php/Vnd/Pkg
    else
      dest = "#{attributes[:prefix]}/#{@target_dir}"
      FileUtils.mkdir_p(dest = "#{pkg.staging_path}/#{dest}")
      FileUtils.cp_r(glob("#{build_path}/*"), dest)
      @name_prefix = "php-"
    end

    # add dependencies
    pkg.name = "#{@name_prefix}#{@name}"
    pkg.version = version_rewrite(pkg.version, typ)
    if typ == :rpm
      pkg.provides += ["php-composer(#{@in_bundle}) = #{pkg.version}"]
    end
    unless attributes[:no_depends_given?]
      pkg.dependencies += dependency_map(:require, typ)
      pkg.provides +=    dependency_map(:provide,  typ)
      pkg.replaces +=    dependency_map(:replace,  typ)
      pkg.conflicts +=   dependency_map(:conflict, typ)
      #pkg.suggests +=    dependency_map(:suggest,  typ)  if typ == :deb
    end
    return pkg
  end # def convert



  private

  #-- Dependency and version translation minutae --



  # translate FPM plugin types into short :rpm, :deb, :phar constants
  def package_typ(klass)
    map = { FPM::Package::RPM => :rpm, FPM::Package::Deb => :deb, FPM::Package::Phar => :phar }
    if map.include?(klass)
      return map[klass]
    elsif klass.to_s =~ /::Filter/i
      return :filter
    else
      return :deb
    end
  end

  # rewrite and copy dependency lists from :composer[require*] to pkg. attributes
  def dependency_map(require, typ)
    if attrs[:composer].key? require.to_s
      return  attrs[:composer][require.to_s].collect { |k,v| require_convert(k, v, @name_prefix, typ) }.flatten.compact
    end
    []
  end

  # translate package names and versions
  def require_convert(k_orig, v, prefix, typ)
    # Map of dependency prefixes
    pn = {
      :php => { :phar => "php",     :deb => "php5-common", :rpm => "php(language)" },
      :ext => { :phar => "ext:",    :deb => "php5-",       :rpm => "php-" },
      :lib => { :phar => "sys:lib", :deb => "lib",         :rpm => "lib" },
      :bin => { :phar => "bin:",    :deb => "",            :rpm => "/usr/bin/" }
    }

    # Package names
    case k = k_orig.strip.gsub(/\W+/, "-")
      when /^php(-32bit|-64bit)?|^hhvm|^quercus/
        k = pn[:php][typ]
        @architecture = ($1 == "-32bit") ? "x86" : "amd64" if $1
      when /^(ext|lib|bin)-(\w+)$/
        k = pn[$1.to_sym][typ] + $2
      else
        if (typ == :rpm)
          k = "php-composer(#{k_orig})" # |php-#{k}"
        elsif typ == :deb and attributes[:composer_no_alt_given?]
          k = "#{prefix}#{k}"
        elsif prefix and prefix.length
          k = "php-#{k}|phar-#{k}" # Allow combinations of plain php- and phar- packages
        else
          # :phar targets have unprefixed vnd-pkg names
        end
    end

    # Expand version specifiers
    if v.index("|")
      # Alternatives go unsupported, because Debians pkg-php-tools doesn't handle them either.
      # It's nearly impossible in RPM land anyway, as RPM depresolvers largely don't permit it.
      # (Properly implementing it also would require consolidating FPMs dependency handling first.)
      v = [""]
      #v=v.split("|").map {}
    else
      rx = "[\\d.]+ ([-~]*\\w.*)?"
      v = v.split(",").map {
        |v|
        v = version_rewrite(v, typ)
        case v
          when /^[*]$|^dev-/                          # any version
            ""
          when "self.version"                         # pseudo JSOX
            " = " + @version
          when /^ (=|==|>=|>|>>|<<|<|<=) (#{rx}) $/x  # >= 2.0.x
            " #{$1} #{$2}"
          when /^ (!=|<>) (#{rx}) $/x                 # conflicts
            attrs[:composer]["conflict"][k] = $1
          when /^ ([\d.]+) \.[\*x] ([~-]*.*)? $/x     # Asterisk 1.0.* and 1.0.x
            [" >= #{$1}.0#{$2}", " < #{next_major_ver($1+'.0')}"]
          when /^ \~  (#{rx}) $/x                     # Tilde ~2.0
            [" >= #{$1}", " < #{next_major_ver($1)}"]
          when /^ \^ (#{rx})  $/x                     # Caret ^1.2.3
            [" >= #{$1}", " < #{next_major_ver($1, 0)}"]
          when /^ #{rx} $/x                           # literal 1.0.1
            " = #{v}"
          else
            ""
        end
      } # AND "," version conditions
    end
    
    # Flatten [pkg|name] + [ver|ver|..] chains into fpm dependency list
    return k ? join_deps(k, v) : nil
  end

  # Expand tilde/caret version "1.2" base to "2.0" next major
  def next_major_ver(v, max=-1)
    v = v.split(".")
    v.pop
    if max == -1
      v[-1] = v[-1].to_i + 1   # increase last version part
    else
      v = [(v[0].to_i + 1).to_s, 0, 0]   # increment major version only
    end
    v.map { |i| i.to_s }.join "."
  end
  
  # Treat "name|name" as list input for [>=version,~>version] array
  def join_deps(k, v)
    v.flatten.map {
      |v|
      k.split("|").map {
        |k|
        k + v
      }.join(" | ")
    }
  end
  
  # Normalize version strings to packaging system
  def version_rewrite(v, typ)
    v = v.gsub(/ ^.+ \sAS\s (.+)$ /unix, "\\1").gsub(/ \s+ | ^v(?=\d) | \#[[:xdigit:]]+$ /unix, "")
    case typ
      when :deb
        # http://anonscm.debian.org/cgit/pkg-php/pkg-php-tools.git/tree/share/php/pkgtools/phpcomposer/source.php
        # https://bugs.debian.org/cgi-bin/pkgreport.cgi?which=maint&data=pkg-php-maint%40lists.alioth.debian.org
        v.gsub!(/[-@](dev|patch)(.*)$/i) { "~~" + "$1$2".downcase }
        v.gsub!(/[-@](alpha|beta|RC|stable)(.*)$/i) { "~" + "$1$2".downcase }
      when :rpm
        v.gsub!(/[-@](dev|patch|alpha|beta|RC|stable)(.*)$/) { $1.downcase }
      else
        v.gsub!(/@/, "-")
    end
    return v
  end


  # Collect per-package composer.json infos
  def composer_json_import(json)
    if json.key? "name" and not attributes[:vendor_given?]
      @vendor = json["name"].split(/\W/).first
    end
    if json.key? "version" and not attributes[:version_given?]
      @version = json["version"].sub(/^v/, "")
    end
    if json.key? "description" and not attributes[:description_given?]
      @description = json["description"]
    end
    if json.key? "license" and not attributes[:license_given?]
      @license = [json["license"]].flatten.join(", ")
    end
    if json.key? "homepage" and not attributes[:url_given?]
      @url = json["homepage"]
    end
    if json.key? "authors" and not attributes[:maintainer_given?]
      @maintainer = json["authors"].map{ |v| v.values.join(", ") }.first or nil
    end
    # stash away complete composer struct for possible phar building
    @attrs[:composer] = json
  end

  # Extract package sections from composer.lock file, turn into pkgname→hash
  def parse_lock(fn, in_bundle)
    json = JSON.parse(File.read(fn))
    if !json.key? "packages"
      json["packages"] = []
    end
    if json.key? "packages-dev"
      json["packages"] += json["packages-dev"]
    end
    lock = Hash[  json["packages"].map{ |entry| [entry["name"], entry] }  ]
    unless lock.key? in_bundle
      raise FPM::InvalidPackageConfiguration, "Package name '#{in_bundle}' absent in composer.lock"
    end
    return lock
  end

  # (unused)
  # Add composer.lock package date into per-package composer.json→extra→lock
  def inject_lock(fn, extra)
    json = JSON.parse(File.read(fn))
    json["extra"] ||= {}
    json["extra"]["lock"] = extra
    File.write(fn, JSON.pretty_generate(json))
  end


  # Locate composer binary
  def composer
    (`which composer` or `which composer.phar` or ("php "+`locate composer.phar`)).split("\n").first
  end

  def glob(path)
    ::Dir.glob(path)
  end

end # class ::Composer

